#if defined(__ICC)
#include <mathimf.h>
#else
#if defined(C99)
#include <tgmath.h>
#else
#include <math.h>
#endif 
#endif 
#include "parms_vec.h"
#include "parms_mat.h"
#include "parms_pc.h"
#include "parms_solver_impl.h"

typedef struct dgmres_data {
  int restart;
  int neig;
} *dgmres_data;

#if defined(FORTRAN_CAPS)
#define dgmr_  DGMR
#elif defined(FORTRAN_DOUBLE_UNDERSCORE)
#define dgmr_  dgmr__
#elif !defined(FORTRAN_UNDERSCORE)
#define dgmr_  dgmr
#endif

extern void dgmr_(int *, int *, int *, int *, int *, int *, int *,
		  double *, double *, double *, int *, int *, double
		  *, int *, int *, int *, int *, double *);   

/** 
 * Deflated GMRES.
 * 
 * @param self Solver object.
 * @param y    Right-hand-side vector.
 * @param x    Solution vector.
 * 
 * @return 0 on success.
 */
int parms_dgmres(parms_Solver self, parms_Vec y, parms_Vec x) 
{

  int n, its,im, maxits,ip,iout,ioutEig,ipo,ipi,nvv,neig,icode,i; 
  FLOAT *sol, *b, eps, ro,*vv;
  dgmres_data fdata;
  parms_Mat A;
  parms_PC  pc;
  parms_Map is;
  parms_Vec wk1, wk2;

  /* check the effectness of solver and vector objects */
  PARMS_VALIDSOLVERPTR(self);
  PARMS_VALIDVECPTR(x);
  PARMS_VALIDVECPTR(y);

  /* retrieve information */
  A    = self->A;
  pc   = self->pc;
  is   = A->is;
  eps  = self->tol;
  fdata = (dgmres_data)self->data;
  im   = fdata->restart;
  maxits = self->maxits;
  n    = parms_MapGetLocalSize(is);
  parms_VecGetArray(x, &sol);
  parms_VecGetArray(y, &b);

  neig = fdata->neig;

  /* create working vector */
  parms_VecCreate(&wk1, is); parms_VecCreate(&wk2, is);

  /* check that no inner iterations are made -- otherwise reset */ 
  nvv = n * (im + 1 + neig + 2);
  ip = 0;
  iout = 0;
  ioutEig = 0;
  PARMS_NEWARRAY(vv, nvv);
  icode = 0;
 label100:
  dgmr_(&n, &im, &neig, &ip, &maxits, &iout, &ioutEig, &eps, b, sol,
	&i, &its, vv, &nvv, &icode, &ipo, &ipi, &ro);  
  
  if (icode <= 0) {
    PARMS_FREE(vv);
    parms_VecFree(&wk1);
    parms_VecFree(&wk2);
    self->its = its;
  }
  else if (icode == 2) { /* matrix vector multiply */
    parms_VecAttachArray(wk1, &vv[ipo-1]);
    parms_VecAttachArray(wk2, &vv[ipi-1]);
    parms_MatApply(A, wk1, wk2);
    goto label100;
  }
  else if (icode == 1) {
    /* preconditioning process */
    parms_VecAttachArray(wk1, &vv[ipo-1]);
    parms_VecAttachArray(wk2, &vv[ipi-1]);
    parms_PCSolve(pc, wk1, wk2);
    goto label100;
  }
  else {
    fprintf(stderr, "*** ERROR in dgmr, icode must equal -1,0,1,2 ***\n"); 
    exit(1);
  }
  return 0; 
}

static int dgmres_free(parms_Solver *self)
{
  return 0;
}

static int dgmres_view(parms_Solver self, parms_Viewer v)
{
  return 0;
}

static int dgmres_setksize(parms_Solver self, int restart)
{
  dgmres_data fdata;
  
  fdata = (dgmres_data)self->data;
  fdata->restart = restart;
  return 0;
}

static int dgmres_setneig(parms_Solver self, int neig)
{
  dgmres_data fdata;

  fdata = (dgmres_data)self->data;
  fdata->neig = neig;
  return 0;
}

/** Create a DGMRES solver. 
 *
 *  \param self A parms_Solver object.
 *  \return 0 on success.
 */
int dgmres_create(parms_Solver self)
{
  dgmres_data fdata;

  self->d_vptr->apply = parms_dgmres;
  self->d_vptr->setksize = dgmres_setksize;
  self->d_vptr->setneig  = dgmres_setneig;
  self->o_vptr->dtor  = (o_dtor)dgmres_free;
  self->o_vptr->view  = (o_view)dgmres_view;
  PARMS_NEW(fdata);
  fdata->restart = 30;
  fdata->neig    = 4;
  self->data =fdata;
  return 0;
}
