/*--------------------------------------------------------------------
  fgmres_create    : create the FGMRES solver. 
  fgmres_free      : free the memory for the FGMRES solver. 
  fgmres_setksize  : set the restart size of the FGMRES solver.
  fgmres_view      : dump the FGMRES solver.
  parms_fgmres     : the FGMRES solve function.

  $Id: fgmres.c,v 1.2 2006/12/01 05:10:00 zli Exp $
  ------------------------------------------------------------------*/

#if defined(__ICC)
#include <mathimf.h>
#else
#if defined(C99)
#include <tgmath.h>
#else
#include <math.h>
#endif 
#endif 
#include "parms_vec.h"
#include "parms_mat.h"
#include "parms_pc.h"
#include "parms_solver_impl.h"

typedef struct fgmres_data {
  int restart;
} *fgmres_data;

#define TINY 1.0e-20
int parms_fgmres(parms_Solver self, parms_Vec y, parms_Vec x)
{
    BOOL outflag, intflag;
    int i, i1, j, restart, its;
    int ii, k, k1, maxits, nloc, size;
    FLOAT  **hh, *hhloc, *c, *s, *rs, *buf, *buf_end_hh;
    FLOAT  eps1, tol, ro, t, alpha, gam;
    parms_Vec *vv, *w;
    parms_Mat A;
    parms_PC  pc;
    parms_Map is;
    fgmres_data fdata;

    PARMS_VALIDSOLVERPTR(self);
    PARMS_VALIDVECPTR(x);
    PARMS_VALIDVECPTR(y);
    A  = self->A;
    pc = self->pc;
    is = A->is;
    fdata = (fgmres_data)self->data;
    restart = fdata->restart;
    maxits  = self->maxits;
    tol     = self->tol;
    nloc    = parms_MapGetLocalSize(is);

    /* permutes x and y */
    parms_VecPerm(x); 
    parms_VecPerm(y);

    /* allocate memory for working local arrays */
    PARMS_NEWARRAY(vv, restart+1);
    for (i = 0; i < restart+1; i++) {
      parms_VecCreate(&vv[i], is);
    }
    PARMS_NEWARRAY(w,  restart);
    for (i = 0; i < restart; i++) {
      parms_VecCreate(&w[i], is);
    }

    PARMS_NEWARRAY(hh, restart);

    size = 4*restart+2+restart*(restart+1)+(2*restart+1)*(nloc+1);
    PARMS_NEWARRAY(buf, size);

    c = buf;
    s = buf + restart;
    rs = s + restart;
    hhloc = rs + restart + 1;
    for (i = 0; i < restart; i++) {
      hh[i] = hhloc+(1+i)*(restart+1);
    }

    buf_end_hh = hhloc+(1+restart)*(restart+1);

    for (i = 0; i < restart+1; i++) {
      parms_VecAttachArray(vv[i], buf_end_hh+i*(nloc+1));
    }

    buf_end_hh += (restart+1)*(nloc+1);

    for (i = 0; i < restart; i++) {
      parms_VecAttachArray(w[i], buf_end_hh+i*(nloc+1));
    }

    /* outer loop starts here */
    its = 0;

    outflag = true;
    while(outflag) {
      /* compute vv[0] = A * x */
      parms_MatApply(A, x, vv[0]);

      /* compute residual vector */
      parms_VecAYPX(vv[0], y, -1.0);
      /* compute the norm of the residual */
      parms_VecGetNorm2(vv[0], &ro);
      if(fabs(ro) <= tol) {
	outflag = false;
	break;
      }
      t = 1.0 / ro;
      parms_VecScale(vv[0], t);
      if(its == 0)
	eps1 = tol*ro;

      /* initialize 1-st term of rhs of hessenberg system */
      rs[0] = ro;
      i = -1;
      intflag = true;
      while (intflag) {
	i++;
	its++;
	i1 = i + 1;
	
	/* preconditioning operation w = K^{-1}vv */
	parms_PCSolve(pc, vv[i], w[i]);
	/*	parms_VecAYPX(w[i], vv[i], 1.0); */
	/* compute A*w */
	parms_MatApply(A, w[i], vv[i1]);
	/* classical Gram-Schmidt */
	parms_VecDotArray(vv[i1], i1, vv, hhloc, hh[i]);
	for (j = 0; j < i1; j++) {
	  alpha = -hh[i][j];
	  parms_VecAXPY(vv[i1], vv[j], alpha);
	}
	parms_VecGetNorm2(vv[i1], &t);
	hh[i][i1] = t;
	if (fabs(t) > TINY) {
	  t = 1.0 / t;
	  parms_VecScale(vv[i1], t);
	}
	/* done with classical Gram-Schmidt and Arnoldi step. now
	   update factorization of hh */
	if (i != 0) {
	  for (k = 1; k <= i; k++) {
	    k1 = k - 1;
	    t = hh[i][k1];
	    hh[i][k1] = c[k1]*t + s[k1]*hh[i][k];
	    hh[i][k] = -s[k1]*t + c[k1]*hh[i][k];
	  }
	}
	gam = sqrt(hh[i][i]*hh[i][i] + hh[i][i1]*hh[i][i1]);
	if (fabs(gam) <= TINY) {
	  gam = TINY;
	}
	/* determine-next-plane-rotation */
	c[i] = hh[i][i]/gam;
	s[i] = hh[i][i1]/gam;
	
	rs[i1] = -s[i]*rs[i];
	rs[i] = c[i]*rs[i];
	/* determine res. norm and test for convergence */
	hh[i][i] = c[i]*hh[i][i] + s[i]*hh[i][i1];
	ro = fabs(rs[i1]);
	if ((i+1 >= restart) || (ro <= eps1) || its >=
	    maxits) {  
	  intflag = false;
	}
      }

      /* now compute solution first solve upper triangular system */ 
      rs[i] = rs[i]/hh[i][i];
      for (ii = 2; ii <= i1; ii++) {
	k = i-ii+1;
	k1 = k+1;
	t = rs[k];
	for (j = k1; j <= i; j++) {
	  t = t - hh[j][k]*rs[j];
	}
	rs[k] = t/hh[k][k];
      }
      /* done with back substitution. now form linear combination to
	 get solution */ 
      for (j = 0; j <= i; j++) {
	t = rs[j];
	parms_VecAXPY(x, w[j], t);
      }
      /* test for return */
      if ((ro <= eps1) || (its >= maxits)) {
	outflag = false;
      }
    }

    PARMS_FREE(buf);
    for (i = 0; i < restart+1; i++) {
      parms_VecFree(&vv[i]);
    }
    PARMS_FREE(vv);

    for (i = 0; i < restart; i++) {
      parms_VecFree(&w[i]);
    }
    PARMS_FREE(w);
    PARMS_FREE(hh);

    self->its = its;
    return 0;
}

static int fgmres_free(parms_Solver *self)
{
  return 0;
}

static int fgmres_view(parms_Solver self, parms_Viewer v)
{
  return 0;
}

static int fgmres_setksize(parms_Solver self, int restart)
{
  fgmres_data fdata;
  
  fdata = (fgmres_data)self->data;
  fdata->restart = restart;
  return 0;
}

static int fgmres_setneig(parms_Solver self, int neig)
{
  return 0;
}

/** Create the FGMRES solver. 
 *
 *  \param self A parms_Solver object.
 *  \return 0 on success.
 */
int fgmres_create(parms_Solver self)
{
  fgmres_data fdata;

  self->d_vptr->apply = parms_fgmres;
  self->d_vptr->setksize = fgmres_setksize;
  self->d_vptr->setksize = fgmres_setneig;
  self->o_vptr->dtor  = (o_dtor)fgmres_free;
  self->o_vptr->view  = (o_view)fgmres_view;
  PARMS_NEW(fdata);
  fdata->restart = 30;
  self->data =fdata;
  return 0;
}

