/*--------------------------------------------------------------------
  parms_SolverApply    : solve the linear system of equation 
  parms_SolverCreate   : create a parms_Solver object.
  parms_SolverFree     : free the memory for the parms_Solver object.
  parms_SolverGetIts   : get the iteration counts.
  parms_SolverGetMat   : get the matrix for the linear system solved.
  parms_SolverGetPC    : get the preconditioning matrix.
  parms_SolverSetParam : set the parameters for the solver (maxits,
                         tolerance, etc.)
  parms_SolverSetType  : set the type of the solver. FGMRES and DGMRES
                         are supplied in the package.
  parms_SolverSetup    : set up the parms_Solver object.
  parms_SolverView     : dump the maximum iteration count and the
                         iteration counts

  A code fragment for using solver functions:

  parms_Solver solver;
  parms_Mat A;
  parms_PC  pc;
  parms_Vec rhs, x;

  // create a solver 
  parms_SolverCreate(&solver, A, pc);

  // set the maximum number of iterations  
  parms_SolverSetParam(solver, MAXITS, "300");
  // set the restart size of FGMRES or DGMRES
  parms_SolverSetParam(solver, KSIZE, "60");
  // set the convergence tolerance 
  parms_SolverSetParam(solver, DTOL, "1.0e-6");
  // set the number of eigenvectors
  parms_SolverSetParam(solver, NEIG, "8"); 
  // set the solver 
  parms_SolverSetup(solver);

  // solver the linear system 
  parms_SolverApply(solver, rhs, x);

  // free the memory for solver
  parms_SolverFree(&solver);

  $Id: parms_solver.c,v 1.6 2006/12/01 16:18:20 zli Exp $
 -------------------------------------------------------------------*/
#include <stdlib.h>
#include "parms_vec.h"
#include "parms_viewer.h"
#include "parms_mat_impl.h"
#include "parms_pc_impl.h"
#include "parms_solver_impl.h"

static int parms_Solver_count = 0;

extern int fgmres_create(parms_Solver self);
extern int dgmres_create(parms_Solver self);

typedef int (*SOLVER_CREATE)(parms_Solver self);

static SOLVER_CREATE solver_tab[] = {
  fgmres_create,
  dgmres_create
};

static int solver_view(parms_Solver self, parms_Viewer v)
{
  FILE *fp;

  PARMS_VALIDSOLVERPTR(self);
  parms_ViewerGetFP(v, &fp);
  fprintf(fp, "maxits = %d  its = %d", self->maxits, self->its);
  parms_ViewerStoreFP(v, fp);
  return 0;
}

/** 
 * Create a parms_Solver object.
 * 
 * @param self A pointer to the parms_Solver object created.
 * @param A    The matrix of the linear system.   
 * @param pc   The preconditioner.  
 * 
 * @return 0 on success.
 */
int parms_SolverCreate(parms_Solver *self, parms_Mat A, parms_PC pc)
{
  parms_Solver new_solver;

  PARMS_VALIDMATPTR(A);
  PARMS_VALIDPCPTR(pc);

  PARMS_OBJCREATE(new_solver, (o_dtor)0, (o_view)solver_view);
  new_solver->handler = PARMS_OBJECT_SET_CLASS(PARMS_SOLVER_CLASS) +
    parms_Solver_count++;
  new_solver->d_vptr->apply    = 0;
  new_solver->d_vptr->setksize = 0;
  new_solver->stype     = SOLFGMRES;
  new_solver->issettype = false;
  new_solver->issetup   = false;
  new_solver->isview    = false;
  new_solver->A = A;
  PARMS_OBJADDREF(A);
  new_solver->pc = pc;
  PARMS_OBJADDREF(pc);
  new_solver->maxits = 100;
  new_solver->tol    = 1.0e-6;
  *self = new_solver;
  return 0;
}

/** 
 * Solve the equation \f$Ax = y\f$.
 * 
 * @param self A parms_Solver object.     
 * @param x    The solution vector.       
 * @param y    The right-hand-side vector.
 * 
 * @return 0 on success.
 */
int parms_SolverApply(parms_Solver self, parms_Vec y, parms_Vec x)
{
  PARMS_VALIDSOLVERPTR(self);
  PARMS_VALIDVECPTR(y);
  PARMS_VALIDVECPTR(x);
  if (self->issettype == false) {
	parms_SolverSetType(self, SOLFGMRES);
  }
  return self->d_vptr->apply(self, y, x);
}
  
/** 
 * Set the type of the solver.
 *
 * Only FGMRES solver is available in the package.
 * 
 * @param self   A parms_Solver object.       
 * @param stype  The type of Krylov subspace. 
 *               -SOLFGMRES
 *               -SOLDGMRES
 *               
 * @return 0 on success.
 */
int parms_SolverSetType(parms_Solver self, SOLVERTYPE stype)
{
  PARMS_VALIDSOLVERPTR(self);
  if (self->issettype && self->stype == stype) {
    return 0;
  }
  if (self->issettype) {
    self->o_vptr->dtor((parms_Object *)&self);
  }
  solver_tab[stype](self);
  self->stype      = stype;
  self->issettype  = true;
  return 0;
}
 
/** 
 * Set parameter for the solver.
 * 
 * Set the maximum iteration counts, the restart size of GMRES, and
 * the convergence tolerance.
 * 
 * @param self   A parms_Solver object.
 * @param ptype  The type of parameter.
 *               -MAXITS maximum iteration counts.
 *               -KSIZE  restart size of GMRES.
 *               -DTOL   converence tolerance.
 *               -NEIG   number of eigenvectors.
 * @param param  Parameters for the solver.
 */
void parms_SolverSetParam(parms_Solver self, PARAMTYPE paramtype, char
			  *param)  
{

  PARMS_VALIDSOLVERPTR(self);
  if (self->issettype == false) {
    parms_SolverSetType(self, SOLFGMRES);
  }
  if (paramtype == MAXITS) {
    self->maxits = atoi(param);
  }
  else if (paramtype == KSIZE) {
    self->d_vptr->setksize(self, atoi(param));
  }
  else if (paramtype == DTOL) {
    self->tol = strtod(param, (char **)NULL);
  }
  else if (paramtype == NEIG) {
    self->d_vptr->setneig(self, atoi(param));
  }
}

/** 
 * Free the memory for the parms_Solver object.
 * 
 * @param self A pointer to the parms_Solver object to be freed. 
 * 
 * @return 0 on success.
 */
int parms_SolverFree(parms_Solver *self)
{
  PARMS_VALIDSOLVERPTR(*self);
  PARMS_OBJSUBREF(*self);
  if (PARMS_OBJGETREF(*self) == 0) {
    parms_MatFree(&(*self)->A);
    parms_PCFree(&(*self)->pc);

    (*self)->o_vptr->dtor((parms_Object *)self);
  }
  return 0;
}

/** 
 * Get the iteration counts.
 * 
 * @param self A parms_Solver object.
 * 
 * @return The iteration counts.
 */
int parms_SolverGetIts(parms_Solver self)
{
  PARMS_VALIDSOLVERPTR(self);

  return self->its;
}

/** 
 * Get the matrix of the linear system.
 * 
 * @param self A parms_Solver object.              
 * @param A    A pointer to the matrix returned.   
 * 
 * @return 0 on success.
 */
int parms_SolverGetMat(parms_Solver self, parms_Mat *A)
{
  PARMS_VALIDSOLVERPTR(self);
  if (self->issetup) {
    *A = self->A;
    return 0;
  }
  else {
    *A = NULL;
    return -1;
  }
}

/** 
 * Get the preconditioning matrix.
 * 
 * @param self A parms_Solver object.                    
 * @param PC   A pointer to the preconditioning matrix.  
 * 
 * @return 0 on success.
 */
int SolverGetPC(parms_Solver self, parms_PC *PC)
{
  PARMS_VALIDSOLVERPTR(self);
  if (self->issetup) {
    *PC = self->pc;
    return 0;
  }
  else {
    *PC = NULL;
    return -1;
  }
}

/** 
 * Setup parms_Solver solver self.
 * 
 * @param self A parms_Solver object.
 * 
 * @return 0 on success.
 */
int parms_SolverSetup(parms_Solver self)
{
  self->issetup = true;
  return 0;
}
