/*--------------------------------------------------------------------
  parms_VecAXPY          : y = y + alpha x
  parms_VecAYPX          : y = scalar y + x
  parms_VecAlloc         : allocate memory for a vector object.
  parms_VecAttachArray   : attach the vector to the memory pointed to
                           by array
  parms_VecAttachVec     : attach the vector to another vector.
  parms_VecCreate        : create a vector object.
  parms_VecDOT           : inner product of two vectors.
  parms_VecDotArray      : inner products for the vector to an array
                           of vector object.
  parms_VecFree          : free the memory for the vector.
  parms_VecGetArray      : get values for the local part of the vector 
  parms_VecGetGlobalSize : get the global size of a vector object.
  parms_VecGetLocalSize  : get the local size of a vector object.
  parms_VecGetNorm2      : get the 2-norm of a vector.
  parms_VecPerm          : permute a vector object.
  parms_VecPut           : assign a scalar value to all components of
                           a vector object.
  parms_VecRestoreArray  : restore values to a vector.
  parms_VecScale         : scale the components of a vector.
  parms_VecSetValues     : insert/add values to a vector.
  parms_VecSetup         : set up a vector 
  parms_VecView          : dump a vector

  A code fragment for using vector functions:

  parms_Vec vec;
  parms_Map map;

  // create a vector.
  parms_VecCreate(&vec, map);

  // assume PE0 contains global variables (0, 2, 4, 6, 8), PE1
  // contains global variables (1, 3, 5, 7, 9). Both PE0 and PE1 call
  // the following parms_VecSetValues function. Each PE will pick and
  // insert variables which belong to the local PE according to the
  // partionting information stored in map.

  // insert values the vector
  im = 0;
  values = 10.0;
  parms_VecSetValues(vec, 1, &im, &values, PARMS_INSERT);
  im = 1;
  values = 11.0;
  parms_VecSetValues(vec, 1, &im, &values, PARMS_INSERT);
  ...
  im = 9
  values = 19.0;
  parms_VecSetValues(vec, 1, &im, &values, PARMS_INSERT);

  // set up the vecotr. after that, no more values can be inserted
  // into the vector
  parms_VecSetup(vec);

  // all vector computation functions can be called
  parms_VecGetNorm2(vec, &value);
  parms_VecDot(vec, vec, &value);
  ...
  // free the memory for the vector
  parms_VecFree(&vec);

  $Id: parms_vec.c,v 1.1.1.1 2006/11/24 21:15:31 zli Exp $
  ------------------------------------------------------------------*/
#include "parms_vec_impl.h"

static int parms_Vec_count = 0;

/** 
 * Free the memory for the vector object.
 * 
 * @param self A pointer to the vector object.
 * 
 * @return 0 on success.
 */
static int vec_Free(parms_Vec *self)
{
  PARMS_OBJSUBREF(*self);
  if (PARMS_OBJGETREF(*self) == 0) {
    parms_MapFree(&(*self)->is);
    if ((*self)->isalloc) {
      PARMS_FREE((*self)->array);
    }
    PARMS_OBJFREE(*self);
  }
  return 0;
}

/** 
 * Dump the vector object.
 * 
 * Output the vector to the viewer v.
 * 
 * @param self A vector object.
 * @param v    A viewer object.
 * 
 * @return 0 on success.
 */
static int vec_View(parms_Vec self, parms_Viewer v)
{
  int lsize, i;
  FLOAT *yvec;
  FILE *fp;
  
  parms_VecGetArray(self, &yvec);

  lsize = parms_VecGetLocalSize(self);
  parms_ViewerGetFP(v, &fp);
  fprintf(fp, "=====Dump the content of parms_Vec=====\n");
  if (self->is->isserial) {
    if ((sizeof(self->array) == sizeof(float)) ||
	(sizeof(self->array) == sizeof(double))) {
      for (i = 0; i < lsize; i++) {
	fprintf(fp, "vec(%d) = %12.5g\n", i, yvec[i]);
      }
    }
#ifdef C99
    else if ((sizeof(self->array) == sizeof(float _Complex)) ||
	     (sizeof(self->array) == sizeof(double _Complex))) {
      for (i = 0; i < lsize; i++) {
	fprintf(fp, "vec(%d) = %12.5g%+12.5gI\n", i, creal(yvec[i]),
		cimag(yvec[i])) ;
      }
    }
#endif 
  }
  else {
    if ((sizeof(self->array) == sizeof(float)) ||
	(sizeof(self->array) == sizeof(double))) {
      for (i = 0; i < lsize; i++) {
	fprintf(fp, "vec[%d,%d] =  %12.5g\n",i,
		self->is->lvars[self->is->iperm[i]], yvec[i]);
      }
    }
#ifdef C99
    else if ((sizeof(self->array) == sizeof(float _Complex)) ||
	     (sizeof(self->array) == sizeof(double _Complex))) {
      for (i = 0; i < lsize; i++) {
	fprintf(fp, "vec[%d,%d] = %12.5g%+12.5gI\n", i, 
		self->is->lvars[self->is->iperm[i]], creal(yvec[i]),
		cimag(yvec[i])); 
      }
    }
#endif 
  }

  fprintf(fp, "=======================================\n");
  parms_VecRestoreArray(self, &yvec);
  parms_ViewerStoreFP(v, fp);
  
  return 0;
}

/** 
 * Get a pointer to a contiguous region containing the local part of
 * the distributed vector.
 * 
 * @param self  A vector object.                            
 * @param array A pointer to the local array in vector self.
 * 
 * @return 0 on success.
 */
static int vec_GetArray(parms_Vec self, FLOAT **array)
{
  int lsize;

  if (!self->isalloc && !self->isatt) {
    lsize = parms_VecGetLocalSize(self);
    PARMS_NEWARRAY(self->array, lsize+1);
    self->isalloc = true;
  }

  *array = self->array + FINDEX;
  return 0;
}

/** 
 * Restore values to the parms_Vec object self.
 *
 * After calling this function, array cannot be used.
 * 
 * @param self   A vector object.                                  
 * @param array  A pointer to the array of values to be restored in
 *               the parms_Vec object self.         
 * 
 * @return 0 on success.
 */
static int vec_Restore(parms_Vec self, FLOAT **array)
{
  self->array = *array - FINDEX;
  *array = NULL;
  return 0;
}

/** 
 * Get the local size of parms_Vec self.
 * 
 * @param this A parms_Vec object.
 * 
 * @return The local size of parms_Vec object self.
 */
static int vec_Getlsize(parms_Vec self)
{
  return self->lsize;
}

/** 
 * Get the global size of parms_Vec object this.
 * 
 * @param this A vector object.
 * 
 * @return The global size of parms_Vec object self.
 */
static int vec_Getgsize(parms_Vec self)
{
  return self->gsize;
}

/** 
 * Assign value to each component of the parms_Vec object self.
 * 
 * @param self   A parms_Vec object.         
 * @param value  A scalar value to be set.
 * 
 * @return 0 on success.
 */
static int vec_Put(parms_Vec self, FLOAT value)
{
  int rem, lsize, i;
  FLOAT *yvec;

  lsize = parms_VecGetLocalSize(self);
  parms_VecGetArray(self, &yvec);
  rem = lsize % 8;

  for (i =  0; i < lsize-rem; i += 8) {
    yvec[i]   = value;
    yvec[i+1] = value;
    yvec[i+2] = value;
    yvec[i+3] = value;
    yvec[i+4] = value;
    yvec[i+5] = value;
    yvec[i+6] = value;
    yvec[i+7] = value;
  }
  for (i = lsize-rem; i < lsize; i++) {
    yvec[i] = value;
  }

  parms_VecRestoreArray(self, &yvec);
  return 0;
}

/** 
 * Scale a vector.
 *
 * All components of parms_Vec object self times scalar. 
 * \f$self = scalar \times self\f$.
 *  
 * @param self   A vector object.
 * @param scalar A scalar.      
 * 
 * @return 0 on success.
 */
static int vec_Scale(parms_Vec self, FLOAT scalar) 
{
  int rem, lsize, i;
  FLOAT *yvec;

  lsize = parms_VecGetLocalSize(self);
  parms_VecGetArray(self, &yvec);
#ifdef HAS_BLAS
  i = 1;
  GSCAL(lsize, scalar, yvec, i);
#if 0
#ifdef C99
  else if ((sizeof(yvec) == sizeof(double  _Complex)) &&
	   (sizeof(scalar) == sizeof(double))) { 
    ZDSCAL(lsize, scalar, yvec, i);
  }
  else if ((sizeof(yvec) == sizeof(double _Complex)) &&
	   (sizeof(scalar) == sizeof(double _Complex))) {
    ZSCAL(lsize, scalar, yvec, i);
  }
#endif
#endif 
#else
  rem = lsize % 8;
  for (i =  0; i < lsize-rem; i += 8) {
    yvec[i]   *= scalar;
    yvec[i+1] *= scalar;
    yvec[i+2] *= scalar;
    yvec[i+3] *= scalar;
    yvec[i+4] *= scalar;
    yvec[i+5] *= scalar;
    yvec[i+6] *= scalar;
    yvec[i+7] *= scalar;
  }
  for (i = lsize-rem; i < lsize; i++) {
    yvec[i] *= scalar;
  }
#endif 
  parms_VecRestoreArray(self, &yvec);
  return 0;
}

/** 
 * Insert values to parms_Vec object self.
 * 
 *  A pseudo code from the global point of view:
 *
 *  \f{verbatim}
 *  for (i = 0; i < m; i++) {
 *    self[im[i]] = values[i]; 
 *  }
 *  \f}
 *  
 * @param self   A vector object.                          
 * @param m      The number of variables to be inserted.   
 * @param im     An array of global variable indices.      
 * @param value  An array of values to be inserted to self.s 
 * @param mode   The style of set values:
 *               -ADD    add values to parms_Vec object self. 
 *               -INSERT assign values to parms_Vec object self.
 *               
 * @return 0 on success.
 */
static int vec_Setvalues(parms_Vec self, int m, int *im, FLOAT
			       *values, INSERTMODE mode) 
{
  int lsize, rindex, index, lrindex, *rp, i;
  FLOAT *yvec;

  lsize = parms_VecGetLocalSize(self);
  if (!self->isalloc && !self->isatt) {
    PARMS_NEWARRAY0(self->array, lsize+1);
    self->isalloc = true;
  }

  parms_VecGetArray(self, &yvec);

  if (!self->is->isserial) {
    if (self->isperm) {
      for (i = 0; i < m; i++) {
	rindex = im[i] - self->is->start;
	rp = parms_MapGlobalToLocal(self->is, rindex);
	if ((rp != NULL) && ((index = *rp) < lsize)) {
	  lrindex = self->is->perm[index];
	  if (mode == ADD) {
	    yvec[lrindex] += values[i];
	  }
	  else if (mode == INSERT) {
	    yvec[lrindex] = values[i];
	  }
	}
      }
    }
    else {
      for (i = 0; i < m; i++) {
	rindex = im[i] - self->is->start;
	rp = parms_MapGlobalToLocal(self->is, rindex);
	if ((rp != NULL) && ((lrindex = *rp) < lsize)) {
	  if (mode == ADD) {
	    yvec[lrindex] += values[i];
	  }
	  else if(mode == INSERT) {
	    yvec[lrindex] = values[i];
	  }
	}
      }
    }
  }
  else {
    for (i = 0; i < m; i++) {
      rindex = im[i] - self->is->start;
      if (self->isperm) {
	rindex = self->is->perm[rindex];
      }
      if (mode == ADD) {
	yvec[rindex] += values[i];
      }
      else if(mode == INSERT) {
	yvec[rindex] = values[i];
      }
    }
  }
  parms_VecRestoreArray(self, &yvec);
  return 0;
}

/** 
 * Perform \f$self := scalar \times x + self\f$.
 * 
 * @param self   A vector object.      
 * @param x      Another vector object.
 * @param scalar A scalar.
 * 
 * @return 0 on success.
 */
static int vec_AXPY(parms_Vec self, parms_Vec x, FLOAT scalar)   
{
  int lsize, rem, i;
  FLOAT *yvec, *xvec;

  lsize = parms_VecGetLocalSize(self);
  parms_VecGetArray(self, &yvec);
  parms_VecGetArray(x, &xvec);

#ifdef HAS_BLAS
  i = 1;
  GAXPY(lsize, scalar, xvec, i, yvec, i);
#if 0
#ifdef C99
  else if (sizeof(*yvec) == sizeof(double _Complex)) {
    ZAXPY(lsize, scalar, xvec, i, yvec, i);
  }
#endif
#endif 
#else
  rem = lsize % 8;
  for (i =  0; i < lsize-rem; i += 8) {
    yvec[i]   += scalar * xvec[i];
    yvec[i+1] += scalar * xvec[i+1];
    yvec[i+2] += scalar * xvec[i+2];
    yvec[i+3] += scalar * xvec[i+3];
    yvec[i+4] += scalar * xvec[i+4];
    yvec[i+5] += scalar * xvec[i+5];
    yvec[i+6] += scalar * xvec[i+6];
    yvec[i+7] += scalar * xvec[i+7];
  }
  for (i = lsize-rem; i < lsize; i++) {
    yvec[i] += scalar * xvec[i];
  }
#endif 
  parms_VecRestoreArray(self, &yvec);
  parms_VecRestoreArray(x, &xvec);

  return 0;
}

/** 
 * Perform \f$self = scalar \times self + x\f$.
 * 
 * @param self    A vector object.      
 * @param x 	  Another vector object.
 * @param scalar  A scalar.
 * 
 * @return 0 on success.
 */
static int vec_AYPX(parms_Vec self, parms_Vec x, FLOAT scalar) 
{
  FLOAT *yvec, *xvec;
  REAL  t;
  int lsize, rem, i;

  lsize = parms_VecGetLocalSize(self);
  parms_VecGetArray(self, &yvec);
  parms_VecGetArray(x, &xvec);

#ifdef HAS_BLAS
  i = 1;
  t = 1.0;
  GSCAL(lsize, scalar, yvec, i);
  GAXPY(lsize, t, xvec, i, yvec, i);
#if 0
#ifdef C99
  else if (sizeof(yvec) == sizeof(double _Complex)) {
    double t = 1.0;
    ZSCAL(lsize, scalar, yvec, i);
    ZAXPY(lsize, t, xvec, i, yvec, i);
  }
#endif
#endif  
#else
  rem = lsize % 8;
  for (i =  0; i < lsize-rem; i += 8) {
    yvec[i]   = scalar * yvec[i]   + xvec[i];
    yvec[i+1] = scalar * yvec[i+1] + xvec[i+1];
    yvec[i+2] = scalar * yvec[i+2] + xvec[i+2];
    yvec[i+3] = scalar * yvec[i+3] + xvec[i+3];
    yvec[i+4] = scalar * yvec[i+4] + xvec[i+4];
    yvec[i+5] = scalar * yvec[i+5] + xvec[i+5];
    yvec[i+6] = scalar * yvec[i+6] + xvec[i+6];
    yvec[i+7] = scalar * yvec[i+7] + xvec[i+7];
  }
  for (i = lsize-rem; i < lsize; i++) {
    yvec[i] = scalar * yvec[i] + xvec[i];
  }
#endif 
  parms_VecRestoreArray(self, &yvec);
  parms_VecRestoreArray(x, &xvec);
  return 0;
}


static int vec_LDot(parms_Vec self, parms_Vec x, FLOAT *value)    
{
  int lsize, i, rem;
  FLOAT dot, s1, s2, s3, s4, s5, s6, s7, s8;
  FLOAT *xvec, *yvec;

  lsize = parms_VecGetLocalSize(self);
  parms_VecGetArray(x,    &xvec);
  parms_VecGetArray(self, &yvec);

#ifdef HAS_BLAS
  i = 1;
  *value = GDOT(lsize, yvec, i, xvec, i);
#if 0
  else if (sizeof(yvec) == sizeof(double _Complex)) {
    *value = ZDOTC(lsize, yvec, i, xvec, i);
  }
#endif 
#else 
  rem = lsize % 8;
  if ((sizeof(yvec) == sizeof(float)) || (sizeof(yvec) ==
					  sizeof(double))) { 
    s1 = 0.0; s2 = 0.0; s3 = 0.0; s4 = 0.0;
    s5 = 0.0; s6 = 0.0; s7 = 0.0; s8 = 0.0;
    for (i =  0; i < lsize-rem; i += 8) {
      s1 += yvec[i]   * xvec[i];
      s2 += yvec[i+1] * xvec[i+1];
      s3 += yvec[i+2] * xvec[i+2];
      s4 += yvec[i+3] * xvec[i+3];
      s5 += yvec[i+4] * xvec[i+4];
      s6 += yvec[i+5] * xvec[i+5];
      s7 += yvec[i+6] * xvec[i+6];
      s8 += yvec[i+7] * xvec[i+7];
    }
    dot = s1 + s2 + s3 + s4 + s5 + s6 + s7 + s8;
    for (i = lsize-rem; i < lsize; i++) {
      dot  +=  yvec[i] * xvec[i];
    }
  }
#ifdef C99
  else if (sizeof(*yvec) == sizeof(double _Complex)) { 
    s1 = 0.0; s2 = 0.0; s3 = 0.0; s4 = 0.0;
    s5 = 0.0; s6 = 0.0; s7 = 0.0; s8 = 0.0;
    for (i =  0; i < lsize-rem; i += 8) {
      s1 += yvec[i]   * conj(xvec[i]);
      s2 += yvec[i+1] * conj(xvec[i+1]);
      s3 += yvec[i+2] * conj(xvec[i+2]);
      s4 += yvec[i+3] * conj(xvec[i+3]);
      s5 += yvec[i+4] * conj(xvec[i+4]);
      s6 += yvec[i+5] * conj(xvec[i+5]);
      s7 += yvec[i+6] * conj(xvec[i+6]);
      s8 += yvec[i+7] * conj(xvec[i+7]);
    }
    dot = s1 + s2 + s3 + s4 + s5 + s6 + s7 + s8;
    for (i = lsize-rem; i < lsize; i++) {
      dot  +=  yvec[i] * conj(xvec[i]);
    }
  }
#endif 

  *value = dot;
#endif 

  parms_VecRestoreArray(self, &yvec);
  parms_VecRestoreArray(x, &xvec);

  return 0;
}

/** 
 * Perform the inner product of two vectors.
 * 
 * If self and x are real vectors, value = self x^{T}. If self
 * and x are complex vectors, value = self \overline{x}^{T}.
 *
 * @param self   A vector object.                  
 * @param x 	 Another vector object.            
 * @param value  The inner product returned.       
 * 
 * @return 0 on success.
 */
static int vec_Dot(parms_Vec self, parms_Vec x, FLOAT *value)
{
  FLOAT *xvec, *yvec;
  MPI_Comm comm;

  parms_VecGetArray(x, &xvec);
  parms_VecGetArray(self, &yvec);
  comm = self->is->comm;

  if (self->is->isserial) {
    vec_LDot(self, x, value);
  }
  else {
    FLOAT ldot;
    vec_LDot(self, x, &ldot);
    if (sizeof(*value) == sizeof(double)) {
      MPI_Allreduce(&ldot, value, 1, MPI_DOUBLE, MPI_SUM, comm);
    }
#ifdef C99
    if (sizeof(*value) == sizeof(double _Complex)) {
      float rel, img, srel, simg;
      rel = creal(ldot);
      img = cimag(ldot);
      MPI_Allreduce(&rel, &srel, 1, MPI_DOUBLE, MPI_SUM, comm);
      MPI_Allreduce(&img, &simg, 1, MPI_DOUBLE, MPI_SUM, comm);
      *value = srel+simg*I;
    }
#endif 
  }
  parms_VecRestoreArray(x, &xvec);
  parms_VecRestoreArray(self, &yvec);

  return 0;
}

/** 
 * Return the 2-norm of the vector.
 * 
 * @param self  A vector object.    
 * @param value The 2-norm returned.
 * 
 * @return 0 on success.
 */
static int vec_Norm2(parms_Vec self, REAL *value)
{
  FLOAT *yvec;
  int lsize;

  parms_VecGetArray(self, &yvec);
  lsize = self->is->lsize;
  if (self->is->isserial) {
#ifdef HAS_BLAS
    int incr = 1;
    *value = GNRM2(lsize, yvec, incr);
#if 0
    else if (sizeof(*yvec) == sizeof(double _Complex)) {
      *value = DZNRM2(lsize, yvec, indcr);
    }
#endif 
#else
    vec_LDot(self, self, value);
    *value = sqrt(*value);
#endif 
  }
  else {
#if FLOAT == double 
    parms_VecDOT(self, self, value);
/* commented by Satish as the following #elif comparision gives compiler error
#elif FLOAT == double _Complex
    parms_VecDOT(self, self, (double _Complex *)value); 
*/
#endif 
    *value = sqrt(*value);
  }

  parms_VecRestoreArray(self, &yvec);
  return 0;
}

/** 
 * Perform the inner product between self and an array of parms_Vec
 * objects.  
 *
 * The pseudo code:
 *
 *  \f{verbatim}
 *  for (i = 0; i < n; i++) {
 *    result[i] = self * vecarray[i];
 *  }
 *  \f}
 *
 * @param self       A vector object.                              
 * @param n 	     The size of vecarray.                         
 * @param vecarray   An array of vector objects.                   
 * @param aux 	     An auxiliary array.                           
 * @param result     An array of size n to store inner products.   
 * 
 * @return 0 on success.
 */
static int vec_DotArray(parms_Vec self, int n, parms_Vec
			*vecarray, FLOAT *aux, FLOAT *result)
{
  int i;
  MPI_Comm comm;


  comm = self->is->comm;
  if (self->is->isserial) {
    for (i = 0; i < n; i++) {
      vec_LDot(self, vecarray[i], &result[i]);
    }
  }
  else {
    for (i = 0; i < n; i++) {
      vec_LDot(self, vecarray[i], &aux[i]);
    }
    if (sizeof(*aux) == sizeof(double)) {
      MPI_Allreduce(aux, result, n, MPI_DOUBLE, MPI_SUM, comm);
    }
#ifdef C99
    else if (sizeof(*aux) == sizeof(double _Complex)) {
      double *rel, *img, *srel, *simg;
      PARMS_NEWARRAY(rel, n);
      PARMS_NEWARRAY(img, n);
      PARMS_NEWARRAY(srel, n);
      PARMS_NEWARRAY(simg, n);
      for (i = 0; i < n; i++) {
	result[i] = creal(aux[i]);
	img[i] = cimag(aux[i]);
      }
      MPI_Allreduce(rel, srel, n, MPI_DOUBLE, MPI_SUM, comm);
      MPI_Allreduce(img, simg, n, MPI_DOUBLE, MPI_SUM, comm);
      for (i = 0; i < n; i++) {
	result[i] = srel[i]+simg[i]*I;
      }
      PARMS_FREE(rel);
      PARMS_FREE(img);
      PARMS_FREE(srel);
      PARMS_FREE(simg);
    }
#endif 
  }
  return 0;
}

/** 
 * Set up the vector object self.
 *
 * This function must be called before the following parms_Vec
 * computation functions: parms_VecGetNorm2, parms_VecGetArray,
 * parms_VecRestoreArray, parms_VecScale, parms_VecAXPY,
 * parms_VecAYPX, parms_VecDOT, parms_VecDotArray.
 *  
 * @param self A vector object.
 * 
 * @return 0 on success.
 */
static int vec_Setup(parms_Vec self)
{
  self->issetup = true;
  return 0;
}

/** 
 * Attach a vector self to another vector x.
 *
 * Attach vector self to a chunk of vector x with beginning index
 * start. 
 *
 * self[0] \f$\rightarrow\f$  x[start] 
 *
 * self[1] \f$\rightarrow\f$  x[start+1]
 *
 * \f$\cdots\f$
 *
 * self[lsize] \f$\rightarrow\f$  x[start+lsize-1]
 *
 * @param self  A vector object.           
 * @param start The start index of x.      
 * @param x 	The attached vector object.
 * 
 * @return 0 on success.
 */
static int vec_AttachVec(parms_Vec self, int start, parms_Vec x)
{
  FLOAT *xvec;

  PARMS_VALIDVECPTR(self);
  PARMS_VALIDVECPTR(x);
  parms_VecGetArray(x, &xvec);
  if (self->isalloc) {
    PARMS_FREE(self->array);
    self->isalloc = false;
  }

  self->isatt = true;
  self->array = &xvec[start-1];
  return 0;
}

/** 
 * Attach vector self to the memory pointed to by array 
 * 
 * @param self  A vector                             
 * @param array The memory to which self is attached.
 * 
 * @return 0 on success.
 */
static int vec_AttachArray(parms_Vec self, FLOAT *array)
{
  if (self->isalloc) {
    PARMS_FREE(self->array);
    self->isalloc = false;
  }

  self->isatt = true;
  self->array = array - 1;
  return 0;
}

static struct parms_Vec_ops vec_vtbl = {
  vec_Getlsize,
  vec_Getgsize,
  vec_GetArray,
  vec_Restore,
  vec_Norm2,
  vec_Put,
  vec_Scale,
  vec_Setvalues,
  vec_AXPY,
  vec_AYPX,
  vec_Dot,
  vec_DotArray,
  vec_AttachVec,
  vec_AttachArray,
  vec_Setup
};

int parms_VecCreate(parms_Vec *self, parms_Map map)
{
  parms_Vec new_vec;

  PARMS_VALIDMAPPTR(map);
  PARMS_OBJCREATE(new_vec, (o_dtor)vec_Free, (o_view)vec_View);
  new_vec->handler = PARMS_OBJECT_SET_CLASS(PARMS_VEC_CLASS) +
    parms_Vec_count++; 
  PARMS_MEMCPY(new_vec->d_vptr, &vec_vtbl, 1);
  new_vec->gsize   = parms_MapGetGlobalSize(map);
  new_vec->lsize   = parms_MapGetLocalSize(map);
  new_vec->issetup = false;
  if (map->ispermalloc) {
    new_vec->isperm  = map->isperm;
  }
  new_vec->is = map;
  PARMS_OBJADDREF(map);
  new_vec->isalloc = false;
  new_vec->isatt   = false;
  *self = new_vec;
  return 0;
}

int parms_VecFree(parms_Vec *self)
{
  PARMS_VALIDVECPTR(*self);
  return (*self)->o_vptr->dtor((parms_Object *)self);
}

int parms_VecView(parms_Vec self, parms_Viewer v)
{
  PARMS_VALIDVECPTR(self);
  PARMS_VALIDVIEWERPTR(v);
  return self->o_vptr->view((parms_Object)self, v);
}

int parms_VecGetLocalSize(parms_Vec self)
{
  PARMS_VALIDVECPTR(self);
  return self->d_vptr->getlsize(self);
}

int parms_VecGetGlobalSize(parms_Vec self)
{
  PARMS_VALIDVECPTR(self);
  return self->d_vptr->getgsize(self);
}

int parms_VecGetArray(parms_Vec self, FLOAT **array)
{
  PARMS_VALIDVECPTR(self);
  return self->d_vptr->getarray(self, array);
}

int parms_VecRestoreArray(parms_Vec self, FLOAT **array)
{
  PARMS_VALIDVECPTR(self);
  return self->d_vptr->restore(self, array);
}

int parms_VecGetNorm2(parms_Vec self, REAL *value)
{
  PARMS_VALIDVECPTR(self);
  return self->d_vptr->norm2(self, value);
}

int parms_VecScale(parms_Vec self, FLOAT scalar)
{
  PARMS_VALIDVECPTR(self);
  return self->d_vptr->scale(self, scalar);
}

int parms_VecSetup(parms_Vec self)
{
  PARMS_VALIDVECPTR(self);
  return self->d_vptr->setup(self);
}

int parms_VecPut(parms_Vec self, FLOAT value)
{
  PARMS_VALIDVECPTR(self);
  return self->d_vptr->put(self, value);
}

int parms_VecSetValues(parms_Vec self, int m, int *im, FLOAT *values,
		       INSERTMODE mode) 
     
{
  PARMS_VALIDVECPTR(self);
  return self->d_vptr->setvalues(self, m, im,  values, mode);
}

int parms_VecAXPY(parms_Vec self, parms_Vec x, FLOAT scalar)
{
  PARMS_VALIDVECPTR(self);
  PARMS_VALIDVECPTR(x);
  return self->d_vptr->axpy(self, x, scalar);
}

int parms_VecAYPX(parms_Vec self, parms_Vec x, FLOAT scalar)
{
  PARMS_VALIDVECPTR(self);
  PARMS_VALIDVECPTR(x);
  return self->d_vptr->aypx(self, x, scalar);
}

int parms_VecDOT(parms_Vec self, parms_Vec x, FLOAT *value)
{
  PARMS_VALIDVECPTR(self);
  PARMS_VALIDVECPTR(x);
  return self->d_vptr->dot(self, x, value);
}

int parms_VecDotArray(parms_Vec self, int n, parms_Vec *vecarray,
		      FLOAT *aux, FLOAT *result)  
{
  PARMS_VALIDVECPTR(self);
  return self->d_vptr->dotarray(self, n, vecarray, aux, result);
}

int parms_VecAttachVec(parms_Vec self, int start, parms_Vec x)
{
  PARMS_VALIDVECPTR(self);
  PARMS_VALIDVECPTR(x);

  return self->d_vptr->attvec(self, start, x);
}

int parms_VecAttachArray(parms_Vec self, FLOAT *array)
{
  PARMS_VALIDVECPTR(self);
  return self->d_vptr->attarray(self, array);
}

int parms_VecPerm(parms_Vec self)
{
  parms_Map    is;
  int          *perm;
  int          size, i, k;
  FLOAT        *xvec, *newvec;
  
  PARMS_VALIDVECPTR(self);
  size = parms_VecGetLocalSize(self);
  is = self->is;
  if (is->isperm && (self->isperm == false)) { 
    self->isperm = is->isperm;
    perm = is->perm;
    parms_VecGetArray(self, &xvec);
    PARMS_NEWARRAY(newvec, size+1);
    for (i = 0; i < size; i++) {
      k = perm[i];
      newvec[k+1] = xvec[i];
    }
    xvec--;
    PARMS_FREE(xvec);
    xvec = newvec+1;
    parms_VecRestoreArray(self, &xvec);
  }
  return 0;
}


int parms_VecPermAux(parms_Vec self, FLOAT *aux)
{
  parms_Map    is;
  int          *perm;
  int          size, i, k;
  FLOAT        *xvec;
  
  PARMS_VALIDVECPTR(self);
  size = parms_VecGetLocalSize(self);
  is = self->is;
  perm = is->perm;
  parms_VecGetArray(self, &xvec);
  for (i = 0; i < size; i++) {
    k = perm[i];
    aux[k] = xvec[i];
  }
  parms_VecRestoreArray(self, &xvec);

  return 0;
}

int parms_VecInvPermAux(parms_Vec self, FLOAT *aux)
{
  parms_Map    is;
  int          *perm;
  int          size, i, k;
  FLOAT        *xvec, *newvec;
  
  PARMS_VALIDVECPTR(self);
  size = parms_VecGetLocalSize(self);
  is = self->is;

  perm = is->iperm;
  parms_VecGetArray(self, &xvec);
  for (i = 0; i < size; i++) {
    k = perm[i];
    aux[k] = xvec[i];
  }
  parms_VecRestoreArray(self, &xvec);
  return 0;
}


int parms_VecAlloc(parms_Vec self)
{
  int lsize;

  PARMS_VALIDVECPTR(self);
  lsize = parms_VecGetLocalSize(self);
  if (lsize) {
    self->isalloc = true;
    PARMS_NEWARRAY0(self->array, lsize+1);
    return 0;
  }
  else {
    return -1;
  }
}
